#include "u8g.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <util/delay.h>
#include <avr/io.h>


#define EMPTY   0
#define LEFT    1
#define RIGHT   2
#define UP      3
#define DOWN    4
#define LOST    7
#define WON     8
#define NEITHER 9

#define SIZE    4

#define F_CPU   16000000

#define ANIMATION_STEPS 2


u8g_t u8g;

static int8_t started;
static int8_t finished;

static char values[18][7] = {"", "   2", "   4", "   8",
                            "  16", "  32", "  64", " 128",
                            " 256", " 512", "1024", "2048",
                            "4096",  "8192", "16384", "32768",
                            "65536", "131072"};

static uint8_t motion;
static uint8_t direction;
static uint8_t move_i, move_j;
static uint8_t distance[SIZE][SIZE];
uint8_t grid[SIZE][SIZE];

int count = 0;

void init_buttons(void) {

    DDRB = 0xff;
    PORTB = 0x00;

    // UP
    DDRB &= ~(1 << PB0);
    PORTB |= (1 << PB0);

    // DOWN
    DDRB &= ~(1 << PB1);
    PORTB |= (1 << PB1);

    // LEFT
    DDRB &= ~(1 << PB2);
    PORTB |= (1 << PB2);

    // RIGHT
    DDRB &= ~(1 << PB3);
    PORTB |= (1 << PB3);

    //RESET
    DDRB &= ~(1 << PB4);
    PORTB |= (1 << PB4);

}

void init_lcd(void) {

    u8g_Init8Bit(&u8g, &u8g_dev_st7920_128x64_4x_8bit,
            PN(0, 0), PN(0, 1), PN(0, 2), PN(0, 3),
            PN(0, 4), PN(0, 5), PN(0, 6), PN(0, 7), // DATA PORTS
            PN(3, 1),    // ENABLE
            U8G_PIN_NONE,
            U8G_PIN_NONE,
            PN(3, 3),    // RS / DI
            PN(3, 2),    // RW
            PN(3, 4));   // RESET
}



void new_block(void)
{
    uint8_t new_value, i, j;


    uint8_t r = rand() % 256;
    if (r > 192)
        new_value = 2;
    else
        new_value = 1;

    do
    {
        srand(count * rand() + 2048);
        count++;
        i = rand() % SIZE;
        srand(count + rand() + 1024);
        count++;
        j = rand() % SIZE;
    }
    while (grid[i][j] != EMPTY);

    grid[i][j] = new_value;
}
void init(void)
{
    memset(grid, 0, SIZE * SIZE * sizeof(uint8_t));
    new_block();
    new_block();
}

void start(void)
{
    started = 1;
    finished = 0;
    srand(count * rand() + 2048);
    count++;
    init();
}

void begin_move(uint8_t type)
{
    motion = 1;
    direction = type;
    switch (type)
    {
        case LEFT:
            move_i = 0;
            move_j = -1;
            break;
        case RIGHT:
            move_i = 0;
            move_j = 1;
            break;
        case UP:
            move_i = -1;
            move_j = 0;
            break;
        case DOWN:
            move_i = 1;
            move_j = 0;
            break;
    }
}




void merge_line(uint8_t i, uint8_t distance[SIZE])
{
    uint8_t c = 0;
    uint8_t j = 0;
    uint8_t merged = 0;

    memset(distance, 0, SIZE);

    for (j = 0; j < SIZE; j++)
    {

        c = j;
        if (grid[i][j] == EMPTY)
        {
            while (grid[i][c] == EMPTY && c < SIZE)
                c++;

            if (c >= SIZE)
                break;

            grid[i][j] = grid[i][c];
            grid[i][c] = EMPTY;

            distance[c] = c - j - 1;
        }


        if (j > 0 && merged == 0 && grid[i][j - 1] == grid[i][j])
        {
            merged = 1;
            grid[i][j - 1]++;
            grid[i][j] = EMPTY;
            j--;
            distance[c]++;
        }
        else
            merged = 0;
    }
}

void move_left(void){

    uint8_t i;
    memset(distance, 0, SIZE * SIZE * sizeof(uint8_t));
    for (i = 0; i < SIZE; i++)
        merge_line(i, distance[i]);
}


/* rotate the matrix */
void rotate(uint8_t m[SIZE][SIZE], uint16_t angle)
{
    uint8_t i, j, t;
    uint8_t new_m[SIZE][SIZE];

    if (angle == 180)
    {
         for (i = 0; i < SIZE; i++)
            for (j = 0; j<SIZE/2; j++)
            {
                t = m[i][j];
                m[i][j] = m[i][SIZE - j - 1 ];
                m[i][SIZE - j - 1] = t;
            }
       return;

    }
    if (angle == 90)
    {
        for (i = 0; i < SIZE; i++)
            for (j = 0; j < SIZE; j++)
                new_m[i][j] = m[j][SIZE - i - 1];
    }


    if (angle == 270)
    {
        for (i = 0; i < SIZE; i++)
            for (j = 0; j < SIZE; j++)
                new_m[i][j] = m[SIZE - j - 1][i];
    }

    memcpy(m, new_m, SIZE * SIZE * sizeof(uint8_t));
}

void move(uint8_t type, uint8_t distance[SIZE][SIZE])
{

    switch (type)
    {
        case LEFT :

            move_left();
            break;

        case RIGHT:

            rotate(grid,180);
            move_left();
            rotate(grid,180);
            rotate(distance,180);
            break;

        case DOWN:

            rotate(grid,270);
            move_left();
            rotate(grid,90);
            rotate(distance,90);
            break;

        case UP:

            rotate(grid,90);
            move_left();
            rotate(grid,270);
            rotate(distance,270);
            break;
    }

}

uint8_t test(uint8_t type, uint8_t distance[SIZE][SIZE])
{
    uint8_t temp[SIZE][SIZE];
    memcpy(temp, grid, SIZE * SIZE * sizeof(uint8_t));

    move(type, distance);

    uint8_t res = memcmp(temp, grid, SIZE * SIZE * sizeof(uint8_t));
    memcpy(grid, temp, SIZE * SIZE * sizeof(uint8_t));

    return res;
}
void check_buttons(void) {

     // UP
    if (!(PINB & (1 << PB0))) {
        if (started == 1) {
            if (motion == 0 && test(UP, distance))
                begin_move(UP);
        }

        if (finished == 1) {
            started = 3;
            if (motion == 0 && test(UP, distance))
                begin_move(UP);
        }


    }

    // DOWN
    if (!(PINB & (1 << PB1))) {
         if (started == 1) {
            if (motion == 0 && test(DOWN, distance))
                begin_move(DOWN);

         }
         if (finished == 1) {
            started = 3;
           if (motion == 0 && test(DOWN, distance))
                begin_move(DOWN);
        }

    }

    // LEFT
    if (!(PINB & (1 << PB2))) {
         if (started == 1) {
            if (motion == 0 && test(LEFT, distance))
                begin_move(LEFT);

         }
         if (finished == 1) {
            started = 3;
            if (motion == 0 && test(LEFT, distance))
                begin_move(LEFT);
        }

    }

    // RIGHT
    if (!(PINB & (1 << PB3))) {
        if (started == 1) {
            if (motion == 0 && test(RIGHT, distance))
                begin_move(RIGHT);

        }
        if (finished == 1) {
            started = 3;
           if (motion == 0 && test(RIGHT, distance))
                begin_move(RIGHT);
        }



    }

    // RESET
    if (!(PINB & (1 << PB4))) {
        start();
    }
    _delay_ms(10);

}


// evaluates the board
uint8_t evaluate(void)
{
    uint8_t i, j;
    uint8_t dummy[SIZE][SIZE];

    if (!(test(LEFT, dummy)  ||
          test(RIGHT, dummy) ||
          test(UP, dummy)    ||
          test(DOWN, dummy)))
        return LOST;

    for (i = 0; i < SIZE; i++)
        for (j = 0; j < SIZE; j++)
            if (grid[i][j] == 11)
                return WON;

    return NEITHER;
}


void draw(void)
{
    int i, j;

    switch (started)
    {
        case 0:
            u8g_SetDefaultForegroundColor(&u8g);
            u8g_SetFont(&u8g, u8g_font_helvB10);
            u8g_DrawStr(&u8g, 20, 25, "Press button");
            u8g_DrawStr(&u8g, 40, 45, "to start");
            break;

        case -1:
            u8g_SetDefaultForegroundColor(&u8g);
            u8g_SetFont(&u8g, u8g_font_helvB10);
            u8g_DrawStr(&u8g, 25, 15, "YOU WON!");
            u8g_DrawStr(&u8g, 20, 35, "Press button");
            u8g_DrawStr(&u8g, 35, 55, "to start");
            break;

        case -2:
            u8g_SetDefaultForegroundColor(&u8g);
            u8g_SetFont(&u8g, u8g_font_helvB10);
            u8g_DrawStr(&u8g, 25, 15, "YOU LOST!");
            u8g_DrawStr(&u8g, 20, 35, "Press button");
            u8g_DrawStr(&u8g, 35, 55, "to start");
            break;

        default:
            u8g_SetFont(&u8g, u8g_font_helvR08);

            for (i = 0; i < SIZE; i++)
                for (j = 0; j < SIZE; j++)
                    if (grid[i][j] > 0)
                    {
                        u8g_SetDefaultForegroundColor(&u8g);
                        u8g_DrawBox(&u8g,
                                    j * 32 + 1 + distance[i][j] * motion * move_j * (32 / ANIMATION_STEPS),
                                    i * 16 + 1 + distance[i][j] * motion * move_i * (16 / ANIMATION_STEPS),
                                    30, 14);
                        u8g_SetDefaultBackgroundColor(&u8g);
                        u8g_DrawStr(&u8g,
                                    j * 32 + 4 + distance[i][j] * motion * move_j * (32 / ANIMATION_STEPS),
                                    i * 16 + 12 + distance[i][j] * motion * move_i * (16 / ANIMATION_STEPS),
                                    values[grid[i][j]]);
                    }
            break;
    }
}

int main(void)
{



    init_buttons();
    init_lcd();

    started = 0;
    finished = 0;
    u8g_SetRot180(&u8g);

    while(1)
    {
        check_buttons();
        if (started == 1)
        {

            if (evaluate() == WON)
            {
                started = -1;
                finished = 1;
             }
            else if (evaluate() == LOST)
                started = -2;
        }

        u8g_FirstPage(&u8g);
        do
        {
            draw();
        }
        while (u8g_NextPage(&u8g));
        _delay_ms(25);
        if (motion > 0)
        {
            motion = (motion + 1) % ANIMATION_STEPS;
            if (motion == 0)
            {
                move(direction, distance);
                new_block();
            }
        }
    }

}

